#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import copy
import dataclasses
import functools
from dataclasses import Field
from logging import getLogger
from typing import Callable, Union, Self, Mapping

import sqlalchemy
import sqlalchemy.orm
from dynamic_types.class_name import prepare_class_name
from dynamic_types.create import create_dataclass
from no_value import NoValue
from type_cast import get_dataclass_field_name_to_field, to_dataclass

from entity_read.entity import Entity
from entity_read.entity_field import get_sqlalchemy_metadata
from .abstract import AbstractView
from .relation_delete import RelationDeleteView
from .relation_insert import RelationInsertView
from .utils import is_list_relation, get_related_entity_type, get_remote_column
from ..utils.no_default import NoDefault

logger = getLogger(__name__)


@functools.cache
def _create_delta_value_dataclass(base_type: type) -> type:
    return create_dataclass(
        prepare_class_name([base_type], 'delta'), [], {
            "old": base_type,
            "new": base_type
        }
    )


@functools.cache
def _create_delta_relation_dataclass(relation: sqlalchemy.orm.RelationshipProperty) -> type:
    # sorry for this: fast way to break circular dependency
    from .relation_delta import RelationDeltaView
    related_entity_type: type[Entity] = get_related_entity_type(relation)
    return create_dataclass(
        prepare_class_name([related_entity_type], 'delta'), [], {
            "create": list[RelationInsertView[relation]],
            "update": list[RelationDeltaView[relation]],
            "delete": list[RelationDeleteView[relation]],
        }
    )


class _DeltaViewMixin:
# class _DeltaViewMixin(abc.ABC):
    # @classmethod
    # @abc.abstractmethod
    # def get_non_primary_key_names(cls) -> tuple[str, ...]: pass
    #
    # @classmethod
    # @abc.abstractmethod
    # def get_key_to_relation(cls) -> Mapping[str, sqlalchemy.orm.Relationship]: pass

    def has_changes(self) -> bool:
        return any(getattr(self, name) not in (NoValue, NoDefault) for name in self.get_non_primary_key_names()) or \
               any(getattr(self, name) for name in self.get_key_to_relation())

    @classmethod
    def _produce_field_name_to_field(cls, related_entity_type: type[Entity],
                                     relation_remote_column: sqlalchemy.Column = None) -> dict[str, Field | Callable]:
        # sorry for this: fast way to break circular dependency
        from .relation_delta import RelationDeltaView
        field_name_to_field = copy.copy(get_dataclass_field_name_to_field(related_entity_type))
        result = {}
        for field_name, field in field_name_to_field.items():
            field_copy = copy.copy(field)
            sql_alchemy_property = get_sqlalchemy_metadata(field)

            if isinstance(sql_alchemy_property, sqlalchemy.orm.RelationshipProperty):
                relation: sqlalchemy.orm.RelationshipProperty = sql_alchemy_property
                field_copy.default = NoValue
                field_copy.default_factory = dataclasses.MISSING
                # TODO: think about non list case: how should we distinct different view types after serialisation(json)
                field_copy.type = _create_delta_relation_dataclass(relation) if is_list_relation(relation) else Union[
                    RelationInsertView[relation], RelationDeltaView[relation], RelationDeleteView[relation]]

            elif isinstance(sql_alchemy_property, sqlalchemy.Column):
                column: sqlalchemy.Column = sql_alchemy_property
                if relation_remote_column is not None and column == relation_remote_column:
                    field_copy.init = False
                    field_copy.default = NoValue
                else:
                    field_copy.default = NoDefault if column.primary_key else NoValue
                if not column.primary_key:
                    field_copy.type = _create_delta_value_dataclass(field_copy.type)
            else:
                logger.warning(f"ignored attribute: %s, cos %s has unexpected type", field, sql_alchemy_property)

            result[field_name] = field_copy

        result['entity_type'] = related_entity_type
        return result

    # def get_affected_types_and_objects_amount(self) -> tuple[int, int]:
    #     ...

    _base = None

    def get_base(self) -> Entity:
        return self._base

    @classmethod
    def from_comparison(cls, old: Entity, new: Entity) -> Self:
        if isinstance(old, _DeltaViewMixin) or isinstance(new, _DeltaViewMixin):
            raise TypeError(f"DeltaViews cannot be compared")
        old_type = old.entity_type if isinstance(old, AbstractView) else type(old)
        new_type = new.entity_type if isinstance(new, AbstractView) else type(new)
        if old_type != new_type:
            raise TypeError(f"DeltaView requires entities and/or views of same entity type,got: {old_type}, {new_type}")
        entity_type = old_type

        delta_key_to_value = old.get_primary_key_name_to_value()

        for column_name in entity_type.get_non_primary_key_names():
            old_value = getattr(old, column_name)
            new_value = getattr(new, column_name, NoDefault)
            if new_value not in (NoValue, NoDefault) and new_value != old_value:
                if old_value in (NoValue, NoDefault):
                    raise ValueError(f'"old" argument MUST have all attributes present in "new", {column_name=}')
                delta_key_to_value[column_name] = {"old": old_value, "new": new_value}

        for relation_name, relation in entity_type.get_key_to_relation().items():
            old_value = getattr(old, relation_name)
            new_value = getattr(new, relation_name)
            if new_value in (NoValue, NoDefault):
                continue
            if old_value in (NoValue, NoDefault):
                raise ValueError(f'"old" argument MUST have all attributes present in "new", {relation_name=}')
            if not relation.uselist:
                old_value = [old_value]
                new_value = [new_value]

            insert_views, delta_views, delete_views = _compare_values(old_value, new_value, relation)
            delta_key_to_value[relation_name] = {"create": insert_views, "update": delta_views, "delete": delete_views}

        result = to_dataclass(delta_key_to_value, cls[entity_type])
        result._base = old
        return result


def _compare_values(
        left_values: list[Entity], right_values: list[Entity], relation
) -> tuple[list[RelationInsertView], list[_DeltaViewMixin], list[RelationDeleteView]]:
    from entity_read.view.relation_delta import RelationDeltaView
    related_entity_type: type[Entity] = get_related_entity_type(relation)
    remote_column = get_remote_column(relation)

    non_relational_primary_key_names: list[str] = [col.key for col in related_entity_type.get_primary_key_columns() if col != remote_column]
    left_primary_keys_to_value: dict[tuple, Entity] = {tuple(getattr(val, name) for name in non_relational_primary_key_names): val for val in left_values}

    right_values_without_primary_keys: list[Entity] = []
    right_primary_keys_to_value: dict[tuple, Entity] = {}
    for val in right_values:
        primary_key_values = tuple(getattr(val, name) for name in non_relational_primary_key_names)
        if None in primary_key_values or NoValue in primary_key_values or NoDefault in primary_key_values:
            right_values_without_primary_keys.append(val)
        else:
            right_primary_keys_to_value[primary_key_values] = val

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    required_but_missing_right_values: list[Entity] = right_values_without_primary_keys + [
        val for keys, val in right_primary_keys_to_value.items() if keys not in left_primary_keys_to_value]

    existent_but_unused_left_values = []
    updated_values = []
    for keys, left_value in left_primary_keys_to_value.items():
        if right_value := right_primary_keys_to_value.get(keys):
            if delta := RelationDeltaView[relation].from_comparison(left_value, right_value):
                updated_values.append(delta)
        else:
            existent_but_unused_left_values.append(left_value)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    identifier_column_names = [col.name for col in related_entity_type.get_identifier_columns() if col != remote_column]
    _optimize_created_updated_deleted(identifier_column_names, right_values_without_primary_keys,
        required_but_missing_right_values, updated_values, existent_but_unused_left_values)

    insert_views = [to_dataclass(val, RelationInsertView[relation]) for val in required_but_missing_right_values]
    delta_views = [to_dataclass(val, RelationDeltaView[relation]) for val in updated_values]
    delete_views = [to_dataclass(val, RelationDeleteView[relation]) for val in existent_but_unused_left_values]
    return insert_views, delta_views, delete_views


def _optimize_created_updated_deleted(
        identifier_column_names, right_values_without_primary_keys,
        required_but_missing_right_values, updated_values, existent_but_unused_left_values
) -> None:  # updated in place
    identifier_to_required_but_missing_relation_values = {}
    for required in right_values_without_primary_keys:
        identifier: tuple = tuple(getattr(required, name) for name in identifier_column_names)
        identifier_to_required_but_missing_relation_values.setdefault(identifier, []).append(required)

    for existent in existent_but_unused_left_values[:]:  # copy, coz is changed in iteration
        identifier: tuple = tuple(getattr(existent, name) for name in identifier_column_names)
        candidates = identifier_to_required_but_missing_relation_values.get(identifier, [])
        if not candidates:
            continue
        best_candidate = next(iter(candidates))
        from entity_read.view.entity_delta import EntityDeltaView
        best_candidate_delta = EntityDeltaView.from_comparison(existent, best_candidate)
        # TODO: implement best candidate selection
        # best_candidate = None
        # best_candidate_delta = None
        # for candidate in candidates:
        #     candidate_delta = EntityDeltaView.from_comparison(existent, candidate)
        #     if best_candidate_delta is None:
        #         best_candidate_delta, best_candidate = candidate_delta, candidate
        #     else:
        #         candidate_delta_complexity: tuple[int, int] = candidate_delta.get_affected_types_and_objects_amount()
        #         # TODO: use cache here somehow
        #         best_candidate_delta_complexity: tuple[int, int] = best_candidate_delta.get_types_and_objects_amount()
        #         if candidate_delta_complexity < best_candidate_delta_complexity:
        #             best_candidate_delta, best_candidate = candidate_delta, candidate

        candidates.remove(best_candidate)
        if best_candidate_delta.has_changes():
            updated_values.append(best_candidate_delta)
        required_but_missing_right_values.remove(best_candidate)
        existent_but_unused_left_values.remove(existent)


#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
import functools
import inspect
import typing
from logging import getLogger
from typing import Callable, Optional

import sqlalchemy
import sqlalchemy.orm
from sqlalchemy import ForeignKey, Column
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy_tools.entity_helpers.sqlalchemy_base import sqlalchemy_mapper_registry
from more_itertools import first

from .defs import SA_DATACLASS_METADATA_KEY
from .ensure_tablename import ensure_tablename, set_table_schema
from .move import relation_to_descend
from .proxy import EntityProxy, ProxyStep
from ..entity import Entity

default_registry = sqlalchemy_mapper_registry
logger = getLogger(__name__)


def _check_fields_metadata(dataclass_entity_type: type[Entity]) -> None:
    for field in dataclass_entity_type.get_fields():
        if SA_DATACLASS_METADATA_KEY not in field.metadata:
            module = inspect.getmodule(dataclass_entity_type)
            target_line = None
            with open(module.__file__) as f:
                for line in f.readlines():
                    if f' {field.name}:' in line:
                        target_line = line.strip()
            raise ValueError(
                f"Entity '{dataclass_entity_type.__name__}' has field '{field.name}' without sqlalchemy metadata. "
                f"Table fields should be inited as '<name>: <type> = entity_field(<db_type>)', "
                f"relations - as '<name>: list[<remote_entity_type>] = entity_relation_field(<remote_entity_type>)'"
                + (f", got: {target_line!r}" if target_line else '')
            )


@typing.dataclass_transform(field_specifiers=(dataclasses.Field, dataclasses.field))
@functools.cache
def sqlalchemy_dataclass_entity(
        registry_or_class: sqlalchemy.orm.registry = default_registry,
        table_schema: str = None, table_name: str = None,
):
    registry = default_registry
    class_ = None
    if isinstance(registry_or_class, sqlalchemy.orm.registry):
        registry = registry_or_class
    elif isinstance(registry_or_class, type):
        class_ = registry_or_class

    @typing.dataclass_transform(field_specifiers=(dataclasses.Field, dataclasses.field))
    def inner(entity_type: type[Entity]) -> type[Entity]:
        entity_type.__sa_dataclass_metadata_key__ = SA_DATACLASS_METADATA_KEY
        ensure_tablename(entity_type, table_name)
        set_table_schema(entity_type, table_schema)

        dataclass_entity_type = dataclasses.dataclass(entity_type)
        _check_fields_metadata(dataclass_entity_type)

        mapped_entity = registry.mapped(dataclass_entity_type)
        mapper = mapped_entity.__mapper__
        assert len(mapper.tables) == 1
        table = first(mapper.tables)
        _table_to_mapper[table] = mapper.entity

        for key in dir(mapped_entity):
            if key.startswith("_"):
                continue
            value = getattr(mapped_entity, key)
            if not isinstance(value, InstrumentedAttribute):
                continue

            value_property = value.property
            if isinstance(value_property, sqlalchemy.orm.relationships.Relationship):
                remote_class = value_property.entity.entity
                remote_column_to_local = {
                    remote_column.name: local_column.name
                    for local_column, remote_column in value_property.local_remote_pairs
                }
                # descension
                descension_proxy = EntityProxy(remote_class, [ProxyStep(key, remote_column_to_local)])
                relation_to_descend[value_property] = make_descend(descension_proxy)
                # value.descend = make_descend(descension_proxy)
            elif isinstance(value_property, sqlalchemy.orm.properties.ColumnProperty):
                assert len(value_property.columns) == 1
                column: Column = value_property.columns[0]
                if len(column.foreign_keys) > 1:
                    logger.warning(f"NOT IMPLEMENTED ascend for composite foreign keys, skip {column.key}")
                    continue
                if len(column.foreign_keys):
                    foreign_key: ForeignKey = first(column.foreign_keys)
                    column.ascend = make_ascend(foreign_key)
            else:
                raise TypeError(f"unexpected type: {type(value_property)}")

        return mapped_entity

    if class_:
        return inner(class_)
    return inner


_table_to_mapper = {}


def make_descend(proxy: EntityProxy) -> Callable[[], EntityProxy]:
    def inner(alias: Optional[str] = None):
        return proxy.label(alias=alias)
    return inner


def make_ascend(foreign_key: ForeignKey) -> Callable[[], EntityProxy]:
    def inner(alias: Optional[str]):
        local_column_name = foreign_key.parent.name
        referred_column_name = foreign_key.column.name
        steps = [ProxyStep(
            local_column_to_remote={referred_column_name: local_column_name},
            name=local_column_name, alias=alias, reverse=True
        )]

        referred_table = foreign_key.constraint.referred_table
        referred_class = _table_to_mapper[referred_table]
        return EntityProxy(referred_class, steps, alias)
    return inner



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field
from typing import ClassVar, Any, TYPE_CHECKING

from entity_read.sql.atoms import Function, Selectable

if TYPE_CHECKING:
    from ..lower import LowerSelector


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionGetSubqueryLength(Function):
    type_: str = field(default=int, init=False)
    args: tuple[Selectable[list]]
    code: ClassVar[str] = 'len'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'jsonb_array_length({args_sql[0]})'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionExists(Function):
    type_: str = field(default=int, init=False)
    args: tuple['LowerSelector']
    code: ClassVar[str] = 'exists'
    keep_subquery: ClassVar[bool] = True

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'EXISTS ({args_sql[0]})'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionNotExists(Function):
    type_: str = field(default=int, init=False)
    args: tuple['LowerSelector']
    code: ClassVar[str] = 'not_exists'
    keep_subquery: ClassVar[bool] = True

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'NOT EXISTS ({args_sql[0]})'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionCastToStr(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'to_str'

    def _to_sql(self, args_sql: list) -> str:
        return f"{args_sql[0]}::TEXT"


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionSub(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'sub'

    def _to_sql(self, args_sql: list) -> str:
        return f"{args_sql[0]} - {args_sql[1]}"


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionDiv(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'div'

    def _to_sql(self, args_sql: list) -> str:
        return f"{args_sql[0]} / {args_sql[1]}"


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionCastToDate(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = "to_date"

    def _to_sql(self, args_sql: list) -> str:
        return f"TO_DATE({args_sql[0]}, {args_sql[1]})"



###  Договоренности по именованию json-функций следующие
###  Все названия из постгри переводим в свой регистр, а в названиях с path, использующих jsonpath,
###  заменяем path на jsonpath. Для операторов, не имеющих имени построим его на описании, сохраняя уникальность, тогда:
###  -> FunctionJsonExtract
###  ->> FunctionJsonExtractText
###  #> FunctionJsonExtractPath
###  #>> FunctionJsonExtractPathText
###  @> FunctionJsonContains
###  <@ FunctionJsonContainedBy
###  ? FunctionJsonHasKey
###  ?| FunctionJsonHasAnyKey
###  ?& FunctionJsonHasAllKeys
###  || FunctionJsonConcat
###  - FunctionJsonDelete
###  #-  FunctionJsonDeletePath
###  @? FunctionJsonJsonpathExists
###  @@ FunctionJsonJsonpathMatch

### Описание json-операторов в документации postgres: https://www.postgresql.org/docs/9.5/functions-json.html
### Функции jpath: https://www.postgresql.org/docs/current/functions-json.html#FUNCTIONS-SQLJSON-CHECK-EXPRESSIONS
from dataclasses import dataclass, field
from typing import ClassVar, Any

from entity_read.sql.atoms import Function, Selectable


@dataclass(kw_only=True, frozen=True)
class FunctionJsonExtractText(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'json_extract_text'

    def _to_sql(self, args_sql: list) -> str:
        return f"{args_sql[0]} ->> {args_sql[1]}"


@dataclass(kw_only=True, frozen=True)
class FunctionJsonHasKey(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'json_has_key'

    def _to_sql(self, args_sql: list):
        return f"{args_sql[0]} ? {args_sql[1]}"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import dataclass
from typing import TypeVar, Generic, Type

from entity_read.sql.atoms.node import Node

T = TypeVar('T')


@dataclass(kw_only=True, frozen=True, slots=True)
class Selectable(Generic[T], Node, abc.ABC):
    type_: Type[T]
    is_nullable: bool = False

    @abc.abstractmethod
    def get_name(self) -> str:
        ...
